import { describe, it, expect, beforeAll } from "@jest/globals";
const crypto = require("crypto");
import StatusesService from "../src/services/StatusesService";
import {
  RECIPIENT_CLASS_NAME,
  SENDER_CLASS_NAME,
} from "../src/services/StatusesService";
import { SenderStatusObject, arrayBufferToBase64 } from "../src";

describe("StatusesService", () => {
  let statusesService;

  beforeAll(() => {
    statusesService = new StatusesService();
  });

  it("encodes/decodes sender status object", () => {
    const textEncoder = new TextEncoder(); // always utf-8
    const stringBytes = textEncoder.encode("test bytes");
    const stringBytesBase64 = arrayBufferToBase64(stringBytes);

    const senderStatusObject: SenderStatusObject = {
      signatureTime: new Date().getTime(),
      signatureTimeSignature: stringBytesBase64,
      messageSignature: stringBytesBase64,
      attachmentsSignatures: {
        attachment_1: {
          list: [
            { value: stringBytesBase64, preSignHashAlgorithm: "test" },
            { value: stringBytesBase64, preSignHashAlgorithm: "test" },
          ],
        },
        attachment_2: {
          list: [{ value: stringBytesBase64, preSignHashAlgorithm: "test" }],
        },
      },
      messageSignatureHashingAlgorithm:
        StatusesService.MIME_SIGNATURE_HASHING_ALGORITHMS.PARTS,
      senderSystem: StatusesService.EMAIL_VENDORS.GMAIL,
      pseudoPlainPartsSignature: "hhg=",
    };

    const senderStatusObjectToCompare = {
      signatureTime: senderStatusObject.signatureTime,
      signatureTimeSignature: stringBytesBase64,
      messageSignature: stringBytesBase64,
      attachmentsSignatures: {
        attachment_1: {
          list: [
            { value: stringBytesBase64, preSignHashAlgorithm: "test" },
            { value: stringBytesBase64, preSignHashAlgorithm: "test" },
          ],
        },
        attachment_2: {
          list: [{ value: stringBytesBase64, preSignHashAlgorithm: "test" }],
        },
      },
      messageSignatureHashingAlgorithm:
        StatusesService.MIME_SIGNATURE_HASHING_ALGORITHMS.PARTS,
      senderSystem: StatusesService.EMAIL_VENDORS.GMAIL,
      pseudoPlainPartsSignature: "hhg=",
    };

    const statusObjectBytes = statusesService.encodeSenderStatusObject(
      senderStatusObject
    );

    const decodedStatusObject = statusesService.decodeSenderStatusObject(
      statusObjectBytes
    );

    console.log(decodedStatusObject);

    expect(decodedStatusObject).toEqual(senderStatusObjectToCompare);
  });

  it("encodes/decodes recipient status data", () => {
    const textEncoder = new TextEncoder(); // always utf-8
    const stringBytes = textEncoder.encode("test bytes");
    const stringBytesbase64 = arrayBufferToBase64(stringBytes);

    const recipientStatusData = {
      creationDate: new Date().getTime(),
      gatewayDomainHash: stringBytes,
      randomToken: stringBytesbase64,
    };

    const recipientStatusDataToCompare = {
      creationDate: recipientStatusData.creationDate,
      gatewayDomainHash: stringBytesbase64,
      randomToken: stringBytesbase64,
    };

    const recipientStatusDataBytes = statusesService.encodeRecipientStatusData(
      recipientStatusData.creationDate,
      recipientStatusData.gatewayDomainHash,
      recipientStatusData.randomToken
    );

    const decodedRecipientStatusObjectData = statusesService.decodeRecipientStatusData(
      recipientStatusDataBytes
    );

    expect(decodedRecipientStatusObjectData).toEqual(
      recipientStatusDataToCompare
    );
  });

  it("encodes/decodes recipient status object", () => {
    const textEncoder = new TextEncoder(); // always utf-8
    const stringBytes = textEncoder.encode("test bytes");
    const stringBytesBase64 = arrayBufferToBase64(stringBytes);

    const recipientStatusObject = {
      iv: stringBytes,
      encryptedContent: stringBytesBase64,
      decryptors: {
        decryptor_1: stringBytes,
        decryptor_2: stringBytesBase64,
      },
    };

    const recipientStatusObjectBytes = statusesService.encodeRecipientStatusObject(
      recipientStatusObject.iv,
      recipientStatusObject.encryptedContent,
      recipientStatusObject.decryptors
    );

    const decodedStatusObject = statusesService.decodeRecipientStatusObject(
      recipientStatusObjectBytes
    );

    const recipientStatusObjectToCompare = {
      iv: stringBytesBase64,
      encryptedContent: stringBytesBase64,
      decryptors: {
        decryptor_1: stringBytesBase64,
        decryptor_2: stringBytesBase64,
      },
    };

    expect(decodedStatusObject).toEqual(recipientStatusObjectToCompare);
  });

  it("encodes/decodes wrapper data", () => {
    const textEncoder = new TextEncoder(); // always utf-8
    const senderBytes = textEncoder.encode("test sender bytes");
    const recipientBytes = textEncoder.encode("test recipient bytes");

    const senderWrappedData = statusesService.encodeSenderWrapperData(
      arrayBufferToBase64(senderBytes)
    );
    const recipientWrappedData = statusesService.encodeRecipientWrapperData(
      recipientBytes
    );

    const senderWrappedDataDecoded = statusesService.decodeWrapperData(
      senderWrappedData
    );
    const recipientWrappedDataDecoded = statusesService.decodeWrapperData(
      recipientWrappedData
    );

    expect(senderWrappedDataDecoded).toEqual({
      data: arrayBufferToBase64(senderBytes),
      version: 1,
      className: SENDER_CLASS_NAME,
    });

    expect(recipientWrappedDataDecoded).toEqual({
      data: arrayBufferToBase64(recipientBytes),
      version: 1,
      className: RECIPIENT_CLASS_NAME,
    });
  });

  it("assembles wrapped sender status object", () => {
    const signatureTime = new Date().getTime();

    const { privateKey, publicKey } = crypto.generateKeyPairSync("rsa", {
      modulusLength: 2048,
    });

    const sign = crypto.createSign("SHA256");
    sign.update(signatureTime.toString(16));
    sign.end();
    const signature = sign.sign(privateKey);

    const pkString = publicKey.export({
      type: "spki",
      format: "pem",
    });

    const senderStatusObject: SenderStatusObject = {
      signatureTime: signatureTime,
      signatureTimeSignature: signature.toString("base64"),
      messageSignature: arrayBufferToBase64(new TextEncoder().encode("test")),
      messageSignatureHashingAlgorithm: "PARTS",
      attachmentsSignatures: {},
      senderSystem: "test",
    };

    const statusObjectBytes = statusesService.encodeSenderStatusObject(
      senderStatusObject
    );

    const wrappedData = statusesService.encodeSenderWrapperData(
      statusObjectBytes
    );
  });

  it("assembles wrapped recipient status object", () => {
    const textEncoder = new TextEncoder(); // always utf-8

    const recipientStatusData = {
      creationDate: new Date().getTime(),
      gatewayDomainHash: textEncoder.encode("gateway domain hash"),
      randomToken: textEncoder.encode("random token"),
    };

    const recipientStatusDataBytes = statusesService.encodeRecipientStatusData(
      recipientStatusData.creationDate,
      recipientStatusData.gatewayDomainHash,
      recipientStatusData.randomToken
    );

    const recipientStatusObject = {
      iv: textEncoder.encode("AES iv"),
      encryptedContent: recipientStatusDataBytes,
      decryptors: {
        decryptor_1: textEncoder.encode("decryptor 1"),
        decryptor_2: textEncoder.encode("decryptor 2"),
      },
    };

    const recipientStatusObjectBytes = statusesService.encodeRecipientStatusObject(
      recipientStatusObject.iv,
      recipientStatusObject.encryptedContent,
      recipientStatusObject.decryptors
    );

    const wrappedData = statusesService.encodeRecipientWrapperData(
      recipientStatusObjectBytes
    );
    //
    // console.log(
    //   "Received status base64",
    //   arrayBufferToBase64(wrappedData.buffer)
    // );
  });
});
