"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BLOCK_DATA_RETRIEVED = exports.MERKLE_TREE_VERIFIED = exports.TRANSACTION_RETRIEVED = void 0;
const index_1 = require("../../index");
const EventEmitter = require("eventemitter2");
const merkletreejs_1 = require("merkletreejs");
const AeternityService_1 = __importDefault(require("../AeternityService"));
const CloudflareService_1 = __importDefault(require("../CloudflareService"));
const VerificationError_1 = __importDefault(require("./VerificationError"));
exports.TRANSACTION_RETRIEVED = "TRANSACTION_RETRIEVED";
exports.MERKLE_TREE_VERIFIED = "MERKLE_TREE_VERIFIED";
exports.BLOCK_DATA_RETRIEVED = "BLOCK_DATA_RETRIEVED";
class VerificationService extends EventEmitter {
    constructor(cdnUrl, cdnBucket, aeternityNodeUrl, aeternityCompilerUrl, aeternityContractBytecode) {
        super();
        this._aeternityService = null;
        this._cloudflareService = null;
        this._aeternityService = new AeternityService_1.default(aeternityNodeUrl, aeternityCompilerUrl, aeternityContractBytecode);
        this._cloudflareService = new CloudflareService_1.default(cdnUrl, cdnBucket);
    }
    getTransaction(hash) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._aeternityService.getTxDataByHash(hash).catch((err) => {
                this.catchVerificationError(err, `Error getting transaction`);
                return null;
            });
        });
    }
    getBlockData(blockHeight) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._aeternityService.getBlockData(blockHeight).catch((e) => {
                this.catchVerificationError(e, `Error retrieving block`);
                return null;
            });
        });
    }
    verifyStatusData(statusData) {
        return __awaiter(this, void 0, void 0, function* () {
            let verificationError;
            let blockData;
            let merkleTreeDetails;
            let txData;
            try {
                txData = yield this.getTransaction(statusData.transactionHash);
                this.emit(exports.TRANSACTION_RETRIEVED, txData);
                const merkleTree = yield this.getMerkleTreeForTransaction(txData);
                const statusVerified = yield this.verifyMerkleTreeNode(statusData.statusRaw, merkleTree);
                merkleTreeDetails = {
                    verified: statusVerified,
                    treeSize: merkleTree.nodes.length,
                    statusPosition: merkleTree.nodes.indexOf(statusData.statusRaw) + 1,
                    rootHash: merkleTree.rootHash,
                };
                this.emit(exports.MERKLE_TREE_VERIFIED, merkleTreeDetails);
                blockData = yield this.getBlockData(txData.blockHeight);
                this.emit(exports.BLOCK_DATA_RETRIEVED, blockData);
            }
            catch (e) {
                if (e instanceof VerificationError_1.default) {
                    verificationError = e.message;
                }
                else {
                    throw e;
                }
            }
            return {
                statusData,
                merkleTreeDetails,
                blockData,
                verificationError,
                txData,
            };
        });
    }
    getMerkleTreeForTransaction(txData) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const txMerkleTreeData = yield this._aeternityService.decodeContractCallData((_a = txData.tx) === null || _a === void 0 ? void 0 : _a.callData);
                const merkleTreeNodes = yield this._cloudflareService.getMerkleTree(txMerkleTreeData.merkleeTreeFileName);
                return {
                    nodes: merkleTreeNodes,
                    rootHash: txMerkleTreeData.rootNodeHash,
                };
            }
            catch (e) {
                this.catchVerificationError(e, `Error extracting Merkle tree from transaction.`);
            }
        });
    }
    verifyMerkleTreeNode(nodeToVerify, merkleTree) {
        return __awaiter(this, void 0, void 0, function* () {
            const leaves = yield Promise.all(merkleTree.nodes.map((x) => __awaiter(this, void 0, void 0, function* () { return Buffer.from(yield index_1.CryptoService.SHA256(x)); })));
            const tree = new merkletreejs_1.MerkleTree(leaves);
            const leaf = yield index_1.CryptoService.SHA256(nodeToVerify);
            const proof = tree.getProof(Buffer.from(leaf), merkleTree.nodes.indexOf(nodeToVerify));
            const verified = tree.verify(proof, Buffer.from(leaf), Buffer.from(merkleTree.rootHash, "base64"));
            if (!verified) {
                throw new VerificationError_1.default(`Merkle tree not verified.`);
            }
            return verified;
        });
    }
    catchVerificationError(error, verificationFailureMessage) {
        if (error.response) {
            throw new VerificationError_1.default(verificationFailureMessage);
        }
        else {
            throw error;
        }
    }
}
VerificationService.TRANSACTION_RETRIEVED = exports.TRANSACTION_RETRIEVED;
VerificationService.MERKLE_TREE_VERIFIED = exports.MERKLE_TREE_VERIFIED;
VerificationService.BLOCK_DATA_RETRIEVED = exports.BLOCK_DATA_RETRIEVED;
exports.default = VerificationService;
